from typing import Tuple, Dict, Union

import numpy as np

LOSConfigType = Dict[str, Union[np.ndarray, Tuple[float, int]]]


class LOSFactory:
    """Factory responsible for registering and book-keeping of LOS configs."""

    def __init__(self) -> None:
        self._configs = {}

    def register_config(self, name: str, components: LOSConfigType) -> None:
        """Initializes and stores a LOS."""

        error_msg = (
            "Line-of-sight config must either be an array, or a tuple with "
            "the format (start, stop, n, geom) where geom is either "
            "'linear' or 'log'"
        )
        config = {}
        for key, value in components.items():
            if isinstance(value, np.ndarray):
                config[key] = value
            elif isinstance(value, tuple):
                try:
                    start, stop, n, geom = value
                except ValueError:
                    raise ValueError(error_msg)
                if geom.lower() == "linear":
                    geom = np.linspace
                elif geom.lower() == "log":
                    geom = np.geomspace
                else:
                    raise ValueError(error_msg)
                config[key] = geom(start, stop, n)
            else:
                raise ValueError(error_msg)
        self._configs[name] = config

    def get_config(self, name: str) -> Dict[str, np.ndarray]:
        """Returns a registered config."""

        config = self._configs.get(name)
        if config is None:
            raise ValueError(
                f"Config {name} is not registered. Available configs are "
                f"{list(self._configs.keys())}"
            )
        return config



from zodipy._model import ModelFactory
from zodipy import emissivities
from zodipy import parameters


models = ModelFactory()

models.register_model(
    name="planck 2013",
    components=(
        "cloud", 
        "band1", 
        "band2", 
        "band3", 
        "ring", 
        "feature"
    ),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2013,
)
models.register_model(
    name="planck 2015",
    components=(
        "cloud",
        "band1",
        "band2",
        "band3",
    ),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2015,
)
models.register_model(
    name="planck 2018",
    components=(
        "cloud",
        "band1",
        "band2",
        "band3",
    ),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2018,
)



from dataclasses import dataclass
from typing import Iterable, Dict, Union, Tuple

from zodipy._components import BaseComponent, Cloud, Band, Ring, Feature
from zodipy._emissivities import Emissivities

ModelParameterType = Dict[str, Dict[str, float]]
EmissivitiesType = Dict[str, Union[Tuple[float], Dict[str, Tuple[float]]]]


@dataclass
class Model:
    """Data class representing an interplanetary dust model."""

    components: Dict[str, BaseComponent]
    emissivities: Emissivities


class ModelFactory:
    """Factory responsible for registring and book-keeping models."""

    def __init__(self) -> None:
        self._models = {}

    def register_model(
        self,
        name: str,
        components: Iterable[str],
        parameters: ModelParameterType,
        emissivities: EmissivitiesType,
    ) -> None:
        """Initializes and stores a model."""

        model = _init_model(components, parameters, emissivities)
        self._models[name] = model

    def get_model(self, name: str) -> Model:
        """Returns a registered model."""

        model = self._models.get(name)
        if model is None:
            raise ValueError(
                f"model {name} is not registered. Available models are "
                f"{list(self._models.keys())}"
            )

        return model


def _init_model(components, parameters, emissivities):
    """Handles mapping of compoents to correct classes."""
    
    initialized_components = {}
    for comp in components:
        if comp.startswith("cloud"):
            comp_type = Cloud
        elif comp.startswith("band"):
            comp_type = Band
        elif comp.startswith("ring"):
            comp_type = Ring
        elif comp.startswith("feature"):
            comp_type = Feature
        initialized_components[comp] = comp_type(**parameters[comp])

    emissivities = Emissivities(**emissivities)

    return Model(
        components=initialized_components, emissivities=emissivities
    )



from typing import Callable

import numpy as np

EmissionCallable = Callable[
    [float, np.ndarray, np.ndarray, np.ndarray, float], np.ndarray
]


def trapezoidal(
    emission_func: EmissionCallable,
    freq: float,
    x_obs: np.ndarray,
    x_earth: np.ndarray,
    x_unit: np.ndarray,
    R: np.ndarray,
    npix: int,
    pixels: np.ndarray,
) -> np.ndarray:
    """Integrates the emission for a component using the trapezoidal method."""

    comp_emission = np.zeros(npix)[pixels]
    emission_prev = emission_func(freq, x_obs, x_earth, x_unit, R[0])
    for i in range(1, len(R)):
        dR = R[i] - R[i - 1]
        emission_cur = emission_func(freq, x_obs, x_earth, x_unit, R[i])
        comp_emission += (emission_prev + emission_cur) * dR / 2
        emission_prev = emission_cur

    return comp_emission



from typing import Optional, Union, Iterable, Dict

import astropy.units as u
import numpy as np

from zodipy._coordinates import get_target_coordinates, to_frame
from zodipy._simulation import get_simulation_strategy
from zodipy.los_configs import LOS_configs
from zodipy.models import models

_EpochsType = Optional[Union[float, Iterable[float], Dict[str, str]]]


class Zodi:
    """The Zodipy interface.

    The Zodiacal emission seen by an observer is highly dependant on the
    specifics of the observation. As an observer moves through the Solar
    System, they will look through different columns of interplanetary
    dust.

    In Zodipy, the simulated Zodiacal emission is a pixel-weighted average
    of all included observations. For each observation, a line-of-sight
    integral is carried out for all observed pixels from the observer
    location. If no hit_counts are provided, the full sky average over all
    observations are returned.

    Parameters
    ----------
    observer
        The observer. Defaults to 'L2'.
    epochs
        The observeration times given as a single epoch, or a list of epochs
        in JD or MJD format, or a dictionary defining a range of times and
        dates; the range dictionary has to be of the form
        {'start':'YYYY-MM-DD [HH:MM:SS]', 'stop':'YYYY-MM-DD [HH:MM:SS]',
        'step':'n[y|d|h|m|s]'}. If no epochs are provided, the current time
        is used in UTC.
    hit_counts
        The number of times each pixel is hit during each observation.
    model
        The Interplanetary dust model used in the simulation. Available
        options are 'planck 2013', 'planck 2015', and 'planck 2018'.
        Defaults to 'planck 2018'.
    line_of_sight_config
        The line-of-sight configuration used in the simulation.
        Available options are: 'default', and 'high', and 'fast'. Defaults
        to 'default'.
    """

    def __init__(
        self,
        observer: str = "L2",
        epochs: Optional[_EpochsType] = None,
        hit_counts: Optional[Iterable[np.ndarray]] = None,
        model: str = "planck 2018",
        line_of_sight_config: str = "default",
    ) -> None:

        model = models.get_model(model)
        line_of_sight_config = LOS_configs.get_config(line_of_sight_config)
        observer_locations = get_target_coordinates(observer, epochs)
        earth_locations = get_target_coordinates("earth", epochs)

        self._simulation_strategy = get_simulation_strategy(
            model, line_of_sight_config, observer_locations, earth_locations, hit_counts
        )

    def get_emission(
        self,
        nside: int,
        freq: Union[float, u.Quantity],
        coord: str = "G",
        return_comps: bool = False,
    ) -> np.ndarray:
        """Simulates the Zodiacal emission in units of MJy/sr.

        Parameters
        ----------
        nside
            HEALPIX map resolution parameter.
        freq
            Frequency at which to evaluate the Zodiacal emission. The
            frequency should be in units of GHz unless an
            `astropy.units.Quantity` object is passed for which it only
            needs to be compatible with Hz.
        coord
            Coordinate frame of the output map. Available options are:
            'E', 'C', or 'G'. Defaults to 'G'.
        return_comps
            If True, the emission of each component in the model is
            returned separatly in the first dim of the output array.
            Defaults to False.

        Returns
        -------
        emission
            Simulated Zodiacal emission in units of MJy/sr.
        """

        if isinstance(freq, u.Quantity):
            freq = freq.to("GHz").value

        emission = self._simulation_strategy.simulate(nside, freq)
        emission = to_frame(emission, coord)

        return emission if return_comps else emission.sum(axis=0)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict
import warnings

import healpy as hp
import numpy as np

from zodipy._integration import trapezoidal
from zodipy._model import Model


@dataclass
class SimulationStrategy(ABC):
    """Base class representing a simulation strategy.

    Attributes
    ----------
    model
        Interplanetary dust model with initialized componentents and
        corresponding emissivities.
    line_of_sight_config
        Dictionary mapping a line_of_sight_config to each component.
    observer_locations
        The location(s) of the observer.
    earth_location
        The location(s) of the Earth.
    hit_counts
        The number of times each pixel is hit during each observation.
    """

    model: Model
    line_of_sight_config: Dict[str, np.ndarray]
    observer_locations: np.ndarray
    earth_locations: np.ndarray
    hit_counts: np.ndarray

    @abstractmethod
    def simulate(self, nside: int, freq: float) -> np.ndarray:
        """Simulates and returns the Zodiacal emission.

        The emission is computed for a given nside and frequency and
        outputted in units of MJy/sr.

        Parameters
        ----------
        nside
            HEALPIX map resolution parameter.
        freq
            Frequency in GHz for which to evaluate the emission.

        Returns
        -------
        emission
            Simulated Zodiacal emission.
        """


@dataclass
class PixelWeightedMeanStrategy(SimulationStrategy):
    """Currently the only implemented simulation strategy.

    This strategy returns the pixel weighted mean of n observations.
    """

    def simulate(self, nside: int, freq: float) -> np.ndarray:
        """See base class for a description."""

        components = self.model.components
        emissivities = self.model.emissivities
        X_observer = self.observer_locations
        X_earth = self.earth_locations
        hit_counts = self.hit_counts

        npix = hp.nside2npix(nside)
        if hit_counts is None:
            hits = np.ones(npix)
            hit_counts = np.asarray([hits for _ in range(len(X_observer))])
        elif hp.get_nside(hit_counts) != nside:
            hit_counts = hp.ud_grade(hit_counts, nside, power=-2)

        X_unit = np.asarray(hp.pix2vec(nside, np.arange(npix)))
        emission = np.zeros((len(components), npix))

        for obs_pos, earth_pos, hit_count in zip(X_observer, X_earth, hit_counts):
            observed_pixels = np.flatnonzero(hit_count)
            unit_vectors = X_unit[:, observed_pixels]

            for comp_idx, (comp_name, comp_class) in enumerate(components.items()):
                comp_emissivity = emissivities.get_emissivity(comp_name, freq)
                line_of_sight = self.line_of_sight_config[comp_name]

                integrated_comp_emission = trapezoidal(
                    comp_class.get_emission,
                    freq,
                    obs_pos,
                    earth_pos,
                    unit_vectors,
                    line_of_sight,
                    npix,
                    observed_pixels,
                )

                emission[comp_idx, observed_pixels] += (
                    integrated_comp_emission
                    * comp_emissivity
                    * hit_count[observed_pixels]
                )

        with warnings.catch_warnings():
            # Unobserved pixels will be divided by 0 in the below return
            # statement. This is fine since we want to return unobserved
            # pixels as np.NAN. However, a RuntimeWarning is raised which
            # we silence in this context manager.
            warnings.filterwarnings("ignore", category=RuntimeWarning)

            return emission / hit_counts.sum(axis=0) * 1e20


def get_simulation_strategy(
    model: Model,
    line_of_sight_config: Dict[str, np.ndarray],
    observer_locations: np.ndarray,
    earth_locations: np.ndarray,
    hit_counts: np.ndarray,
) -> SimulationStrategy:
    """Initializes, validates and returns a simulation strategy."""

    number_of_observations = len(observer_locations)
    if number_of_observations != len(earth_locations):
        raise ValueError(
            "The number of observer locations and earth locations must be " "the same"
        )
    if hit_counts is not None:
        hit_counts = np.asarray(hit_counts)
        number_of_hit_counts = 1 if np.ndim(hit_counts) == 1 else len(hit_counts)
        if number_of_hit_counts != number_of_observations:
            raise ValueError(
                f"The number of 'hit_counts' ({number_of_hit_counts}) are "
                "not matching the number of observations "
                f"({number_of_observations})"
            )

    return PixelWeightedMeanStrategy(
        model, line_of_sight_config, observer_locations, earth_locations, hit_counts
    )



from zodipy._los_config import LOSFactory

import numpy as np

EPS = np.finfo(float).eps
RADIAL_CUTOFF = 6


LOS_configs = LOSFactory()

LOS_configs.register_config(
    name="default",
    components={  # "comp name": (start, stop, n, geometry = "linear" | "log")
        "cloud": (EPS, RADIAL_CUTOFF, 250, "linear"),
        "band1": (EPS, RADIAL_CUTOFF, 50, "linear"),
        "band2": (EPS, RADIAL_CUTOFF, 50, "linear"),
        "band3": (EPS, RADIAL_CUTOFF, 50, "linear"),
        "ring": (EPS, 2.25, 50, "linear"),
        "feature": (EPS, 1, 50, "linear"),
    },
)
LOS_configs.register_config(
    name="high",
    components={
        "cloud": (EPS, RADIAL_CUTOFF, 500, "linear"),
        "band1": (EPS, RADIAL_CUTOFF, 500, "linear"),
        "band2": (EPS, RADIAL_CUTOFF, 500, "linear"),
        "band3": (EPS, RADIAL_CUTOFF, 500, "linear"),
        "ring": (EPS, 2.25, 200, "linear"),
        "feature": (EPS, 1, 200, "linear"),
    },
)
LOS_configs.register_config(
    name="fast",
    components={
        "cloud": (EPS, RADIAL_CUTOFF, 25, "linear"),
        "band1": (EPS, RADIAL_CUTOFF, 25, "linear"),
        "band2": (EPS, RADIAL_CUTOFF, 25, "linear"),
        "band3": (EPS, RADIAL_CUTOFF, 25, "linear"),
        "ring": (EPS, 2.25, 25, "linear"),
        "feature": (EPS, 1, 25, "linear"),
    },
)
